"""Models related to samples, like Samples and SampleTypes."""
import json
from collections.abc import Sequence
from typing import List
from typing import Tuple
from warnings import warn

from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import JSONSaveMixin
from pydent.models.field_value_mixins import FieldTypeInterface
from pydent.models.field_value_mixins import FieldValueInterface
from pydent.relationships import HasMany
from pydent.relationships import HasOne


@add_schema
class Sample(FieldValueInterface, ModelBase):
    """A Sample model."""

    fields = dict(
        # sample relationships
        sample_type=HasOne("SampleType"),
        items=HasMany("Item", ref="sample_id"),
        field_values=HasMany(
            "FieldValue", ref="parent_id", additional_args={"parent_class": "Sample"}
        ),
        user=HasOne("User"),
    )

    METATYPE = "sample_type"

    def __init__(
        self,
        name=None,
        project=None,
        description=None,
        sample_type=None,
        sample_type_id=None,
        properties=None,
        field_values=None,
    ):
        """

        :param name:
        :type name:
        :param project:
        :type project:
        :param description:
        :type description:
        :param sample_type_id:
        :type sample_type_id:
        :param properties:
        :type properties:
        """
        super().__init__(
            name=name,
            project=project,
            description=description,
            sample_type_id=sample_type_id,
            sample_type=sample_type,
            field_values=field_values,
            items=None,
        )

        if properties is not None:
            self.update_properties(properties)

    @property
    def identifier(self):
        """Return the identifier used by Aquarium in autocompletes."""
        return "{}: {}".format(self.id, self.name)

    def field_value(self, name):
        """Returns the :class:`FieldValue` associated with the sample by its
        name. If the there is more than one FieldValue with the same name (as
        in field_value arrays), it will return the first FieldValue. See the.

        :meth:`Sample.field_value_array` method.

        :param name: name of the field value
        :type name: str
        :return: the field value
        :rtype: FieldValue
        """
        return self.get_field_value(name, None)

    def field_value_array(self, name):
        return self.get_field_value_array(name, None)

    def _property_accessor(self, fv):
        ft = self.safe_get_field_type(fv)
        if ft:
            if ft.ftype == "sample":
                return fv.sample
            else:
                if ft.ftype == "number":
                    if isinstance(fv.value, str):
                        return json.loads(fv.value)
                return fv.value

    @property
    def properties(self):
        return self._field_value_dictionary(lambda ft: ft.name, self._property_accessor)

    def field_value_dictionary(self):
        return self._field_value_dictionary(lambda ft: ft.name, lambda x: x)

    def update_properties(self, prop_dict):
        """Update the FieldValues properties for this sample.

        :param prop_dict: values to update
        :type pro fp_dict: dict
        :return: self
        :rtype: Sample
        """
        ft_dict = {ft.name: ft for ft in self.get_field_types()}
        for name, val in prop_dict.items():
            ft = ft_dict[name]
            if ft.is_parameter():
                key = "value"
            else:
                key = "sample"
            if issubclass(type(val), Sequence) and ft.array:
                self.set_field_value_array(name, None, [{key: v} for v in val])
            else:
                self.set_field_value(name, None, {key: val})

    def create(self) -> List["Sample"]:
        """Creates this sample in the Aquarium instance if the sample has all
        fields required by the sample type.
        Uses is_savable to check for missing fields.

        .. versionchanged:: 0.1.5a17
            Raises `AquariumModelError` if sample is missing required properties

        :return: the list of saved samples
        :raises AquariumModelError: if required FieldValues are missing.
        """
        self.is_savable(do_raise=True)
        return self.session.utils.create_samples([self])

    def is_savable(self, do_raise: bool = True) -> Tuple[bool, List[str]]:
        """Checks whether this sample has fields required by the sample type.
           If so, the sample can be saved or updated.

        .. versionadded:: 0.1.5a17

        :param do_raise: raise an exception when required field values are missing
        :return: a boolean whether there are errors, and list of error messages
        :raises AquariumModelError: if do_raise is True and required fields are missing
        """
        errors = []
        for k, v in self.properties.items():
            if v is None and self.sample_type.field_type(k).required:
                errors.append("FieldValue '{}' is required.".format(k))
        if do_raise and errors:
            raise AquariumModelError(
                "Cannot update/save due to the following:\n"
                "Sample: id={} name={} ({})\n\t{}".format(
                    self.id, self.name, self.sample_type.name, "\n\t".join(errors)
                )
            )
        return len(errors) == 0, errors

    def save(self):
        """Saves the sample, either by creating a new sample (if id=None) or
        updating the existing sample on the server.

        .. versionchanged:: 0.1.5a17
            Raises `AquariumModelError` if sample is missing required properties
        :return:
        """
        if self.id:
            self.update()
        else:
            self.create()

    def update(self) -> "Sample":
        """Updates the sample on the server.

        .. versionchanged:: 0.1.5a17
            Raises `AquariumModelError` if sample is missing required properties
        :return:
        """
        self.is_savable(do_raise=True)
        for fv in self.field_values:
            fv.reload(fv.save())

        new_fvs = self.field_values
        server_fvs = self.session.FieldValue.where(
            dict(parent_id=self.id, parent_class="Sample")
        )

        to_remove = [
            fv for fv in server_fvs if fv.id not in [_fv.id for _fv in new_fvs]
        ]
        if to_remove:
            warn(
                "Trident tried to save a Sample, but it required FieldValues to be deleted."
            )
            for fv in to_remove:
                fv.parent_id = None
                fv.save()
        self.reload(self.session.utils.json_save("Sample", self.dump()))
        return self

    def merge(self):
        """Merge sample by name. If a sample with the same name and
        sample_type_id is found on the server, update that model and save the
        updated data to the server. Else, create a new sample on the server.

        :param sample: sample to merge.
        :return: True if merged, False otherwise
        """
        try:
            self.save()
        except Exception as e:
            existing = self.session.Sample.find_by_name(self.name)
            if existing:
                if self.sample_type_id == existing.sample_type_id:
                    existing.update_properties(self.properties)
                    existing.description = self.description
                    existing.project = self.project
                    existing.save()
                    self.reload(existing.dump())
                    return True
                else:
                    raise e
            else:
                raise e
        return False

    def available_items(self, object_type_name=None, object_type_id=None):
        query = {"name": object_type_name, "id": object_type_id}
        query = {k: v for k, v in query.items() if v is not None}
        if query == {}:
            return [i for i in self.items if i.location != "deleted"]
        else:
            object_types = self.session.ObjectType.where(query)
            object_type = object_types[0]
            return [
                i
                for i in self.items
                if i.location != "deleted" and i.object_type_id == object_type.id
            ]

    def copy(self):
        """Return a copy of this sample."""
        copied = super().copy()
        copied.anonymize()
        return copied

    def __str__(self):
        return self._to_str("id", "name", "sample_type")


@add_schema
class SampleType(FieldTypeInterface, JSONSaveMixin, ModelBase):
    """A SampleType model."""

    fields = dict(
        samples=HasMany("Sample", "SampleType"),
        field_types=HasMany(
            "FieldType", ref="parent_id", additional_args={"parent_class": "SampleType"}
        ),
        # TODO: operation_type_afts
        # TODO: property_afts
        # TODO: add relationships description
    )

    def field_type(self, name, role=None):
        return super().field_type(name, role)

    @property
    def properties(self):
        props = {}
        for ft in self.field_types:
            if ft.ftype == "sample":
                props[ft.name] = [str(aft) for aft in ft.allowable_field_types]
            else:
                props[ft.name] = ft.ftype
        return props

    def new_sample(self, name, description, project, properties=None):
        if properties is None:
            properties = dict()
        sample = self.session.Sample.new(
            name=name,
            sample_type=self,
            description=description,
            project=project,
            properties=properties,
        )
        return sample

    def _get_update_json(self):
        return self.dump(include=("field_types",))

    def _get_create_json(self):
        return self.dump(include=("field_types",))

    def __str__(self):
        return self._to_str("id", "name")



"""Models related to data associations."""
import json
import os
import shutil
from typing import Union

import requests

from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import JSONDeleteMixin
from pydent.models.crud_mixin import JSONSaveMixin
from pydent.relationships import HasOne
from pydent.relationships import JSON
from pydent.utils import make_async


# TODO: changing the value of a association (and saving it) shouldn't be difficult
class DataAssociatorMixin:
    """Mixin for handling data associations."""

    def _make_association(self, key, value, parent, upload=None):
        association = self.session.DataAssociation.new(
            object=json.dumps({key: value}),
            upload=upload,
            key=key,
            parent_id=parent.id,
            parent_class=parent.__class__.__name__,
        )
        association.parent = parent
        return association

    def associate(self, key, value, upload=None, create_new=False, save=None):
        """Adds a data association with the key and value to this object.

        :param key: Key of the association
        :type key: str
        :param value: a json serializable object
        :type value: dict | str | int | float
        :param upload: optional file to upload
        :type upload: File
        :param create_new: if True (default) will create a new association instead
            of updating the existing association.
        :tuype create_new: bool
        :return: newly created or updated data association
        :rtype: DataAssociation
        """
        if save is None:
            if self.id is None:
                save = False
            else:
                save = True
        if create_new:
            association = None
        else:
            association = self.get_data_association(key)

        if not association:
            if save:
                return self.session.utils.create_data_association(
                    self, key, value, upload=upload
                )
            else:
                association = self._make_association(key, value, self, upload)
                self.append_to_many("data_associations", association)
                return association
        else:
            association.value = value
            association.upload = upload
            association.parent_class = self.__class__.__name__
            association.parent_id = self.id
            association.parent = self
            if save:
                self._try_update_data_association(association)
            return association

    def delete_data_associations(self, key):
        associations = self.get_data_associations(key)
        for a in associations:
            if a.id:
                a.delete()
            self.data_associations.remove(a)

    def _try_update_data_association(self, association):
        upload = association.upload
        if upload and upload.id:
            association.upload_id = upload.id
        if self.id:
            if association.upload and not association.upload.id:
                association.upload.save()
                association.upload_id = association.upload.id
            association.save()
        return association

        # return self.session.utils.create_data_association(
        #     self, key, value, upload=upload
        # )
        # if unique_key:
        #     existing_associations = self.get_data_associations(key)
        #     for association in existing_associations:
        #         association.value = value
        #         association.upload = upload
        #         association.upload_id = upload.id
        #         association.save()
        # else:
        #     self.session.utils.create_data_association(
        #         self, key, value, upload=upload
        #     )

    # def save_data_associations(self):
    #     if self.is_deserialized('data_associations'):
    #

    def associate_file(self, key, value, file, job_id=None):
        """Associate a file.

        :param key: association key
        :type key: str or json
        :param value: association value
        :type value: str or json
        :param file: file to create :class:`Upload`
        :type file: file object
        :param job_id: optional job_id to associate the :class:`Upload`
        :type job_id: int
        :return: new data association
        :rtype: :class:`DataAssociation`
        """
        u = self.session.Upload.new(job_id=job_id, file=file)
        return self.associate(key, value, upload=u)

    def associate_file_from_path(self, key, value, filepath, job_id=None):
        """Associate a file from a filepath.

        :param key: association key
        :type key: str or json
        :param value: association value
        :type value: str or json
        :param filepath: path to file to create :class:`Upload`
        :type filepath: str
        :param job_id: optional job_id to associate the :class:`Upload`
        :type job_id: int
        :return: new data association
        :rtype: :class:`DataAssociation`
        """
        with open(filepath, "rb") as f:
            return self.associate_file(key, value, f, job_id=job_id)

    def get_data_association(self, key: Union[None, str]):
        das = self.get_data_associations(key)
        if das:
            return das[0]

    def get_data_associations(self, key: Union[None, str]):
        das = []
        if self.data_associations:
            for da in self.data_associations:
                if da.key == key:
                    das.append(da)
        return das

    # TODO: DataAssociation - do we really want to have this return either a list or
    #       single value? How are people using this?
    def get(self, key):
        val = []
        for da in self.get_data_associations(key):
            val.append(da.value)
        if len(val) == 1:
            return val[0]
        elif len(val) == 0:
            return None
        return val


class DataAssociationSaveContext:
    def __init__(self, model: DataAssociatorMixin):
        self.model = model
        self.data_associations = []

    def __enter__(self):
        if self.model.is_deserialized("data_associations"):
            self.data_associations = self.model.data_associations

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            return
        for da in self.data_associations:
            da.parent_id = self.model.id
            self.model._try_update_data_association(da)
            assert da.id
        self.model.data_associations = self.data_associations


@add_schema
class DataAssociation(JSONDeleteMixin, JSONSaveMixin, ModelBase):
    """A DataAssociation model."""

    fields = dict(object=JSON(), upload=HasOne("Upload"))

    @property
    def value(self):
        return self.object.get(self.key, None)

    @value.setter
    def value(self, new_value):
        self.object = {self.key: new_value}

    # def save(self):
    #     if self.upload and not self.upload_id:
    #         self.upload.save()
    #         self.upload_id = self.upload_id
    #     super().save()
    #
    #     data_association = self.parent.session.DataAssociation.find(self.id)
    #     if data_association.id not in [da.id for da in self.parent.data_associations]:
    #         self.parent.data_associations.append(data_association)
    #         return data_association
    #     else:
    #         for da in self.parent.data_associations:
    #             if da.id == data_association.id:
    #                 return da

    def save(self):
        if self.parent_id is None:
            raise ValueError("Cannot save DataAssociation. `parent_id` cannot be None")
        if self.parent_class is None:
            raise ValueError(
                "Cannot save DataAssociation. `parent_class` cannot be None"
            )
        super().save(do_reload=True)

    def __str__(self):
        return self._to_str("id", "object")


@add_schema
class Upload(ModelBase):
    """An Upload model."""

    fields = dict(job=HasOne("Job"))

    def __init__(self, job_id=None, file=None):
        """Create a new upload.

        :param job_id: job id to associate the upload to
        :type job_id: int
        :param file: file to upload
        :type file: file object
        """
        super().__init__(job_id=job_id)
        self.file = file

    query_hook = dict(methods=["size", "name", "job"])

    # def _get_uploads_from_job_id(self, job_id):

    def _get_uploads_from_job(self):
        http = self.session._http
        return http.get("krill/uploads?job={}".format(self.job_id))["uploads"]

    def temp_url(self):
        data = self.session.Upload.where({"id": self.id}, methods=["expiring_url"])[
            0
        ].raw
        return data["expiring_url"]

    @staticmethod
    def _download_file_from_url(url, outpath):
        """Downloads a file from a url.

        :param url: url of file
        :type url: str
        :param outpath: filepath of out file
        :type outpath: str
        :return: http response
        :rtype: str
        """
        response = requests.get(url, stream=True)
        with open(outpath, "wb") as out_file:
            shutil.copyfileobj(response.raw, out_file)
        return response.raw

    @staticmethod
    @make_async(1)
    def async_download(uploads, outdir=None, overwrite=True):
        """Asynchronously downloads from list of :class:`Upload` models.

        :param uploads: list of Uploads
        :type uploads: list
        :param outdir: path to output directory to save downloaded files
        :type outdir: str
        :param overwrite: if True, will overwrite existing files
        :type overwrite: bool
        :return: list of filepaths
        :rtype: list
        """
        return Upload._download_files(uploads, outdir, overwrite)

    @staticmethod
    def _download_files(uploads, outdir, overwrite):
        """Downloads uploaded file from list of :class:`Upload` models.

        :param uploads: list of Uploads
        :type uploads: list
        :param outdir: path to output directory to save downloaded files (defaults to
            current directory)
        :type outdir: str
        :param overwrite: if True, will overwrite existing files (default: True)
        :type overwrite: bool
        :return: list of filepaths
        :rtype: list
        """
        filepaths = []
        for upload in uploads:
            filepath = upload.download(outdir=outdir, overwrite=overwrite)
            filepaths.append(filepath)
        return filepaths

    def fetch(self, outdir: str = None, filename: str = None, overwrite: bool = True):
        """Alias for `download`

        :param outdir: path of directory of output file (default is current directory)
        :param outfile: filename of output file (defaults to upload_filename)
        :param overwrite: whether to overwrite file if it already exists
        :return: filepath of the downloaded file
        """
        return self.download(outdir=outdir, filename=filename, overwrite=overwrite)

    def download(
        self, outdir: str = None, filename: str = None, overwrite: bool = True
    ):
        """Downloads the uploaded file to the specified output directory. If no
        output directory is specified, the file will be downloaded to the
        current directory.

        :param outdir: path of directory of output file (default is current directory)
        :param outfile: filename of output file (defaults to upload_filename)
        :param overwrite: whether to overwrite file if it already exists
        :return: filepath of the downloaded file
        """
        if outdir is None:
            outdir = "."
        if filename is None:
            filename = "{}_{}".format(self.id, self.upload_file_name)
        filepath = os.path.join(outdir, filename)
        if not os.path.exists(filepath) or overwrite:
            self._download_file_from_url(self.temp_url(), filepath)
        return filepath

    @property
    def data(self):
        """Return the data associated with the upload."""
        result = requests.get(self.temp_url())
        return result.content

    def create(self):
        """Save the upload to the server."""
        return self.session.utils.create_upload(self)

    def save(self):
        return self.create()
